
###
# Version 2.1 27 Oct 2014
#	: Changes from version 1.0
#		: bug fix: negative z now allowed
#		: initial value for slope: no longer required to use last two data points (see offset=)
#		: characterise.ise: new function to characterise an ise to use when there are no experimental samples
###

###
# Generates initial values for a single ISE model
# Data should be formatted using load.ISE.data
# If there are convergence problems, initial values
# for the non-linear regression may be altered
# by specifying a.init, b.init, and cstar.init;
# sigma.upper places an upper bound on sigma.init;
# stdadd specifies basic model (F) or standard addition model (T)
# offset determines how many points to use to estimate the slope; 
#	offset = 1 uses the last two points,
#	offset = 2 uses the n-2 and nth point, etc.
###
gen.inits.single = function(data, a.init=NA, b.init=NA, cstar.init=NA, 
	logc.limits = c(-8.9, -1.9), sigma.upper=NA, stdadd = F, offset = 1, calibration.only = F) {
	# Generate initial values for calibration data
	x <- 10^data$log10x
	order.x = order(x)
	x = sort(x)
	log10x = log10(x)
	emf <- data$emf[order.x]
	
	# Assignments
	n = length(x)
	b = (emf[n] - emf[n-offset])/(log10x[n] - log10x[n-offset])	
	a = emf[n] - b*log10x[n]
	C = min(max(10^logc.limits[1], 10^((emf[1]-a)/b) - x[1]), 10^logc.limits[2])
	cstar = 10^(log10(C)/10)

	# over-ride if initial values are provided
	if(!is.na(a.init)) { a = a.init  }
	if(!is.na(b.init)) { b = b.init  }
	if(!is.na(cstar.init)) { cstar = cstar.init  }

	Preds = a + b*log10(x + C)
	Resids = emf - Preds
	logsigma.temp = log(sqrt( sum(Resids^2)/(length(Resids) - 3) ))
	logsigma <- logsigma.temp
	sigma = exp(logsigma)
	# If sigma is too large, it may be out of bounds.
	# Use sigma.upper to place an upper limit on sigma
	if(!is.na(sigma.upper)) {
		sigma = min(sigma.upper, exp(logsigma))
	}

	sigma.sq = sigma^2
	tau = 1/sigma.sq

	# Calculate LOD using s/n = 3
  	mu0 <- a + b*log10(cstar^10)
  	mu0.3sig <- mu0 + 3*sigma
  	LOD <- 10^( (mu0.3sig - a)/b ) - cstar^10

	# Generate inital values for experimental data
	# Values with ISEs below S/N=3 are set to E(ISE|x=0) + 3*sigma
   if (!calibration.only)	{
	x.exp = rep(NA, data$M)
	if (stdadd != T) {
		emf.tmp <- data$emf.exp
		emf.tmp2 <- pmax(emf.tmp, mu0.3sig)
		x.tmp <- 10^( (emf.tmp2 - a)/b ) - cstar^10
	}
	if (stdadd == T) {
		delta.emf.tmp = data$delta.emf
		V.add.tmp = data$V.add
		conc.add.tmp = data$conc.add
		V.s.tmp = data$V.s
	#	x.tmp = (V.add.tmp*conc.add.tmp/V.s.tmp)/(10^(delta.emf.tmp/b))
		x.tmp = pmax(10^logc.limits[1],
			(V.add.tmp*conc.add.tmp/V.s.tmp)/(10^(delta.emf.tmp/b)))
	}
		
	# Initial values are between 90% and 100% of the average estimate
	x.exp = x.tmp*runif(1, 0.9, 1)
	log10x.exp = log10(x.exp)
	
	INITS <- list(a=a, b=b, cstar=cstar, logsigma=logsigma)
	init.stretch = 0.01
	inits <- list(a=a*runif(1, 1 - init.stretch, 1 + init.stretch),
				b=b*runif(1, 1 - init.stretch, 1 + init.stretch),
				cstar=cstar*runif(1, 1 - init.stretch, 1 + init.stretch),
				sigma=sigma*runif(1, 1 - init.stretch, 1 + init.stretch),
				log10x.exp=log10x.exp)
   }
   if (calibration.only) {
	INITS <- list(a=a, b=b, cstar=cstar, logsigma=logsigma)
	init.stretch = 0.01
	inits <- list(a=a*runif(1, 1 - init.stretch, 1 + init.stretch),
				b=b*runif(1, 1 - init.stretch, 1 + init.stretch),
				cstar=cstar*runif(1, 1 - init.stretch, 1 + init.stretch),
				sigma=sigma*runif(1, 1 - init.stretch, 1 + init.stretch))
   }
	return(inits)
}	


###
# Similar to gen.inits.single, but for multiple ISEs
# If initial values are specified, the should be vectors with length equal to the number of ISEs
###
gen.inits.multiple <- function(data, a.init= NA, b.init= NA, cstar.init= NA, offset = 1,
	logc.limits = c(-8.9, -1.9), sigma.upper = NA, stdadd = F, calibration.only=F) {
	# Initialise vectors
	a <- rep(NA, data$R)
	b <- rep(NA, data$R)
	cstar <- rep(NA, data$R)
	logsigma <- rep(NA, data$R)
	sigma <- rep(NA, data$R)
	sigma.sq <- rep(NA, data$R)
	tau <- rep(NA, data$R)

	# Generate initial values for calibration data
	for (i in 1:data$R) {
		
		x <- 10^data$log10x[data$ISEID==i]
		order.x = order(x) 			#new
		x = sort(x)  #new
		log10x = log10(x)
		emf <- data$emf[data$ISEID==i][order.x] #new

		# Assignments
		n = length(x)
		b[i] = (emf[n] - emf[n-offset])/(log10x[n] - log10x[n-offset])	
		a[i] = emf[n] - b[i]*log10x[n]
		C = min(max(10^logc.limits[1], 10^((emf[1]-a[i])/b[i]) - x[1]), 10^logc.limits[2])
		cstar[i] = 10^(log10(C)/10)

		# over-ride if initial values are provided
		if(!is.na(a.init[1])) { a[i] = a.init[i]  }
		if(!is.na(b.init[1])) { b[i] = b.init[i]  }
		if(!is.na(cstar.init[1])) { cstar[i] = cstar.init[i]  }

		Preds = a[i] + b[i]*log10(x + C)
		Resids = emf - Preds
		logsigma.temp = log(sqrt( sum(Resids^2)/(length(Resids) - 3) ))

		logsigma[i] <- logsigma.temp
		sigma[i] = exp(logsigma[i])

		# If sigma is too large, it may be out of bounds.
		# Use sigma.upper to place an upper limit on sigma
		if(!is.na(sigma.upper)) {
			sigma[i] = min(sigma.upper, exp(logsigma[i]))
		}
		sigma.sq[i] = sigma[i]^2
		tau[i] = 1/sigma.sq[i]
	}

	# Calculate LOD using s/n = 3
  	mu0 <- a + b*log10(cstar^10)
  	mu0.3sig <- mu0 + 3*sigma*sign(b)
  	LOD <- 10^( (mu0.3sig - a)/b ) - cstar^10

   if (!calibration.only) {
	# Generate inital values for experimental data
	# Values with ISEs below S/N=3 are set to E(ISE|x=0) + 3*sigma
	x.exp = rep(NA, data$M)

	if (stdadd != T) {
		for (i in 1:data$M) {
			emf.tmp <- data$emf.exp[data$xID.exp==i]
			ISEID.tmp <- data$ISEID.exp[data$xID.exp==i]
			emf.tmp2 <- pmax(emf.tmp, mu0.3sig[ISEID.tmp])
			x.tmp <- 10^( (emf.tmp2 - 
				a[ISEID.tmp])/b[ISEID.tmp] ) - 
				cstar[ISEID.tmp]^10
			# Initial values are between 90% and 100% of the average estimate
			# but at least equal to the lower limit for log c
			x.exp[i] = max(10^logc.limits[1], mean(x.tmp))*runif(1, 0.9, 1)
		}
	}

	if (stdadd == T) {
		for (i in 1:data$M) {
			delta.emf.tmp = data$delta.emf[data$xID.exp==i]
			ISEID.tmp <- data$ISEID.exp[data$xID.exp==i]
			V.add.tmp = data$V.add[data$xID.exp==i]
			conc.add.tmp = data$conc.add[data$xID.exp==i]
			V.s.tmp = data$V.s[data$xID.exp==i]
			x.tmp = (V.add.tmp*conc.add.tmp/V.s.tmp)/(10^(delta.emf.tmp/b[ISEID.tmp]))
			# Initial values are between 90% and 100% of the average estimate
			x.exp[i] = mean(x.tmp)*runif(1, 0.9, 1)
		}
	}

	log10x.exp = log10(x.exp)

	INITS <- list(a=a, b=b, cstar=cstar, logsigma=logsigma)
	init.stretch = 0.01
	inits <- list(a=a*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				b=b*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				cstar=cstar*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				sigma=sigma*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				log10x.exp=log10x.exp)
   }
   if (calibration.only) {
	INITS <- list(a=a, b=b, cstar=cstar, logsigma=logsigma)
	init.stretch = 0.01
	inits <- list(a=a*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				b=b*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				cstar=cstar*runif(data$R, 1 - init.stretch, 1 + init.stretch),
				sigma=sigma*runif(data$R, 1 - init.stretch, 1 + init.stretch))
   }
	return(inits)
}


###
# Calculates calibration curve, and calibrates experimental data
# Required inputs are:
#  data:        Data formatted using load.ISE.data
#  model.path:  The directory where the BUGS model is located (e.g. "c:/myisefolder")
#  model.name:  The name of the BUGS model (e.g. "Single ISE model.txt")
#  single.ISE:  T or F, indicating whether the data come from a single ISE or multiple ISEs
#  stdadd:	    T or F (default is F) indicating whether the experimental data is in the standard addition format.
#
# Graphical parameters that may be specified, if desired:
#  plot.limits: The range of the y axis for the sample concentrations plot
#  x.ticks:     The location of tickmarks on the x axis (e.g. x.ticks = c(1, 3, 5) draws ticks at 1, 3, and 5
#  y.ticks:     The location of tickmarks on the y axis
#
# Technical inputs that do not require specification: 
#  burnin:      The number of initial parameter estimates to discard while tuning the system
#  iters:       The total number of iterations
#  chains:      The number of parallel chains to run (useful for diagnostics)
#  ### Note:    The final estimates are based on (iters - burnin)*chains number of simulations
#  a.init:      Initial value passed to gen.inits.single (scalar) or gen.inits.multiple (vector)
#  b.init:      Similar
#  cstar.init:  Similar
#  logc.limits: Upper and lower climits for log(c) (defaults to -8.9 and -1.9)
#  sigma.upper: Upper limit for sigma.init
#  offset:	The initial value for the slope uses the last and the last - offset point (defaults to 1).
#  ### Note:    Default values may be altered if there are convergence problems in the 
#               non-linear regression
#  diagnostic.print:  Prints information used to determine convergence
###

bayesian.calibrate <- function(data, model.path, model.name, single.ISE=NA, Z=NA, temperature = 21, stdadd=F,
	plot.limits = c(-15, 0), x.ticks = NA, y.ticks = NA, label.x = "Sample ID", label.y = expression(paste(log[10], " { ", italic(x), " }" )),
	burnin=25000, iters = 50000, chains=4, thin = 1,
	a.init= NA, b.init=NA, cstar.init=NA, logc.limits = c(-8.9, -1.9), sigma.upper = NA, diagnostic.print=F, offset = 1) {

	# Location of the BUGs model
	ISE.model <- file.path(model.path, model.name)

	# Error checking
	if (is.na(single.ISE)) {
		print("Please specify single.ISE=T or single.ISE=F to indicate a single or multiple ISEs")
		return()
	}

	# Error checking
	if (is.na(Z)) {
		print("Please specify the ionic valence using Z= (e.g. bayesian.calibrate(..., Z = 2,...)")
		return()
	}

	# BUGs model requires at least two samples.  
	# If only one exists, duplicate it.
	flag.datareplicated = 0
	if(data$M==1) {
		data$M=2
		data$emf.exp=rep(data$emf.exp, 2)
		flag.datareplicated = 1
	}

	# Initial values
	cat("\nGenerating initial values...\n")
	if (single.ISE==T) {
		# Number of ISEs = 1
		Num.ISEs = 1
		# Generate initial values
		inits <- rep(list(NA), chains)
		for (i in 1:chains) {
			inits[[i]] <- gen.inits.single(data, a.init=a.init, b.init=b.init,
				cstar.init= cstar.init, logc.limits=logc.limits, sigma.upper=sigma.upper, stdadd = stdadd, offset=offset)
		}
	}
	if (single.ISE==F) {
		# Generate initial values
		inits <- rep(list(NA), chains)
		# Number of ISEs given in data
		Num.ISEs = data$R
		for (i in 1:chains) {
			inits[[i]] <- gen.inits.multiple(data, a.init=a.init, b.init=b.init, 
				cstar.init= cstar.init, logc.limits=logc.limits, sigma.upper= sigma.upper, stdadd = stdadd, offset=offset)
		}
	}
	# Parameters to monitor
	parameters <- c("a", "b", "c", "cstar", "sigma", "log10x.exp")

	# Calculate theoretical slope at mV level
	data$mu.b = 1000*log(10)*8.31433*(temperature + 273.15)/(96487*Z)

	# Call BUGS
	cat("Running Bayesian model...\n\n")
	ISE.Bayes <- bugs(data, inits, parameters, ISE.model,
		n.chains = chains, n.iter = iters, n.burnin=burnin, n.thin=thin,
		program = "openbugs", working.directory = NULL)

	# Print and plot diagnostics
	if(diagnostic.print==T) {
		print(ISE.Bayes, digits=9)
		plot(ISE.Bayes)
	}

  if(single.ISE==T) {
	# Calculate quantiles for x
	log10x.exp <- matrix(NA, nrow=data$M, ncol=3)
	log10x.exp.IQ <- matrix(NA, nrow=data$M, ncol=2)
	colnames(log10x.exp)=c("Estimated concentration", "Lower limit", "Upper limit")
	colnames(log10x.exp.IQ)=c("1st quartile", "3rd quartile")
	for (i in 1:data$M) {
		tmp <- as.vector(ISE.Bayes$sims.array[,,i+5])
		quantile.tmp <- quantile(tmp, c(0.025, 0.5, 0.975, 0.25, 0.75))
		log10x.exp[i,] <- quantile.tmp[c(2,1,3)]
		log10x.exp.IQ[i,] <- quantile.tmp[c(4,5)]
	}
	
	# Other vars
	ahat.coda = as.vector(ISE.Bayes$sims.array[,,1])
	coda.length = length(ahat.coda)
	bhat.coda = as.vector(ISE.Bayes$sims.array[,,2])
	cstarhat.coda = as.vector(ISE.Bayes$sims.array[,,4])
	chat.coda = cstarhat.coda^10
	sigmahat.coda = as.vector(ISE.Bayes$sims.array[,,5])
	ahat = median(ahat.coda)
	bhat = median(bhat.coda)
	cstarhat = median(cstarhat.coda)
	chat = cstarhat^10
	sigmahat = median(sigmahat.coda)

	ahat.lcl = quantile(ahat.coda, 0.025)
	bhat.lcl = quantile(bhat.coda, 0.025)
	cstarhat.lcl = quantile(cstarhat.coda, 0.025)
	chat.lcl = cstarhat.lcl^10
	ahat.ucl = quantile(ahat.coda, 0.975)
	bhat.ucl = quantile(bhat.coda, 0.975)
	cstarhat.ucl = quantile(cstarhat.coda, 0.975)
	chat.ucl = cstarhat.ucl^10
	sigmahat.lcl = quantile(sigmahat.coda, 0.025)
	sigmahat.ucl = quantile(sigmahat.coda, 0.975)


	# Derived variable
	LOD.SN3 = log10( cstarhat^10 *(10^(3*sigmahat/abs(bhat))-1))

	LOD.SN3.coda = 10*log10(cstarhat.coda) + log10(10^(3*sigmahat.coda /abs( bhat.coda ) ) - 1)
	LOD.SN3hat = median(LOD.SN3.coda)
	LOD.SN3.Q1 = quantile(LOD.SN3.coda, 0.25)
	LOD.SN3.Q3 = quantile(LOD.SN3.coda, 0.75)
	LOD.SN3.lcl = quantile(LOD.SN3.coda, 0.025)
	LOD.SN3.ucl = quantile(LOD.SN3.coda, 0.975)

  }

  if (single.ISE == F) {
	# Calculate quantiles for x
	log10x.exp <- matrix(NA, nrow=data$M, ncol=3)
	log10x.exp.IQ <- matrix(NA, nrow=data$M, ncol=2)
	colnames(log10x.exp)=c("Estimated concentration", "Lower limit", "Upper limit")
	colnames(log10x.exp.IQ)=c("1st quartile", "3rd quartile")
	for (i in 1:data$M) {
		tmp = as.vector(ISE.Bayes$sims.array[,,i+5*data$R])
		quantile.tmp <- quantile(tmp, c(0.025, 0.5, 0.975, 0.25, 0.75))
		log10x.exp[i,] <- quantile.tmp[c(2,1,3)]
		log10x.exp.IQ[i,] <- quantile.tmp[c(4,5)]
	}
	
	# Other vars
	coda.length = length(as.vector(ISE.Bayes$sims.array[,,1]))
	ahat.coda = matrix(NA, nrow = coda.length, ncol = data$R)
	bhat.coda <- cstarhat.coda <- chat.coda <- sigmahat.coda <- LOD.SN3.coda <- ahat.coda
	
	ahat = rep(NA, data$R)
	ahat.lcl = rep(NA, data$R)
	ahat.ucl = rep(NA, data$R)
	bhat = rep(NA, data$R)
	bhat.lcl = rep(NA, data$R)
	bhat.ucl = rep(NA, data$R)
	chat = rep(NA, data$R)
	chat.lcl = rep(NA, data$R)
	chat.ucl = rep(NA, data$R)
	cstarhat = rep(NA, data$R)
	cstarhat.lcl = rep(NA, data$R)
	cstarhat.ucl = rep(NA, data$R)
	sigmahat = rep(NA, data$R)
	sigmahat.lcl = rep(NA, data$R)
	sigmahat.ucl = rep(NA, data$R)
	LOD.SN3 = rep(NA, data$R)

	LOD.SN3hat = rep(NA, data$R)
	LOD.SN3.lcl = rep(NA, data$R)
	LOD.SN3.ucl = rep(NA, data$R)
	LOD.SN3.Q1 = rep(NA, data$R)
	LOD.SN3.Q3 = rep(NA, data$R)

	for (j in 1:data$R) {
		ahat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,j])
		bhat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,data$R + j])
		cstarhat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,3*data$R + j])
		sigmahat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,4*data$R + j])
		chat.coda[,j] = cstarhat.coda[,j]^10

		ahat.q = quantile(ahat.coda[,j], c(0.025, 0.5, 0.975))
		ahat[j] = ahat.q[2]
		ahat.lcl[j] = ahat.q[1]
		ahat.ucl[j] = ahat.q[3]
		bhat.q = quantile(bhat.coda[,j], c(0.025, 0.5, 0.975))
		bhat[j] = bhat.q[2]
		bhat.lcl[j] = bhat.q[1]
		bhat.ucl[j] = bhat.q[3]
		chat.q = quantile(chat.coda[,j], c(0.025, 0.5, 0.975))
		chat[j] = chat.q[2]
		chat.lcl[j] = chat.q[1]
		chat.ucl[j] = chat.q[3]
		cstarhat.q = quantile(cstarhat.coda[,j], c(0.025, 0.5, 0.975))
		cstarhat[j] = cstarhat.q[2]
		cstarhat.lcl[j] = cstarhat.q[1]
		cstarhat.ucl[j] = cstarhat.q[3]
		sigmahat.q = quantile(sigmahat.coda[,j], c(0.025, 0.5, 0.975))
		sigmahat[j] = sigmahat.q[2]
		sigmahat.lcl[j] = sigmahat.q[1]
		sigmahat.ucl[j] = sigmahat.q[3]

		# Derived variable
		LOD.SN3[j] = 10*log10( cstarhat[j]) + log10(10^(3*sigmahat[j]/abs(bhat[j]))-1)

		LOD.SN3.coda[,j] = 10*log10( cstarhat.coda[,j]) + log10( 10 ^ (3*sigmahat.coda[,j] / abs( bhat.coda[,j] )) - 1)
		LOD.SN3.q = quantile(LOD.SN3.coda[,j], c(0.025, 0.25, 0.5, 0.75, 0.975))
		LOD.SN3hat[j] = LOD.SN3.q[3]
		LOD.SN3.lcl[j] = LOD.SN3.q[1]
		LOD.SN3.ucl[j] = LOD.SN3.q[5]
		LOD.SN3.Q1[j] = LOD.SN3.q[2]
		LOD.SN3.Q3[j] = LOD.SN3.q[3]
   	}	
   }

	### Plot calibrated samples
	par(mfrow=c(2,2))
	par(mar=c(4,4.0,1,0)+0.1)
	par(mgp=c(2, 0.5, 0), tcl = -0.2)

	SampleID = seq(data$M)
	delta = 0.015*data$M
	S.low = SampleID - delta
	S.high = SampleID + delta
	S.llow = SampleID - 0.5*delta
	S.lhigh = SampleID + 0.5*delta
	plot(rep(SampleID[1],2), c(log10x.exp[1,2], log10x.exp[1,3]), type="l", axes=F,
		xlim=c(min(SampleID), max(SampleID)), ylim=plot.limits, lend = 1, 
		xlab = label.x, ylab= label.y
	)
	box()
	if(!is.na(x.ticks[1])) {
		axis(1, at = x.ticks, cex.axis= 0.8)
	}
	if(is.na(x.ticks[1])) {
		axis(1, at = SampleID, cex.axis = 0.8)
	}
	if(is.na(y.ticks[1])) {
		axis(2, cex.axis= 0.8)
	}
	if(!is.na(y.ticks[1])) {
		axis(2, at = y.ticks, cex.axis= 0.8)
	}

	for (i in 1:data$M) {
		lines(rep(SampleID[i],2), c(log10x.exp[i,2], log10x.exp[i,3]), col="black")
		lines(c(S.llow[i], S.lhigh[i]), rep(log10x.exp[i,2], 2) ,  col="black")
		lines(c(S.llow[i], S.lhigh[i]), rep(log10x.exp[i,3], 2), col="black")
		lines(c(S.low[i], S.high[i]), rep(log10x.exp[i,1], 2), lwd=1, col="black")
		lines(rep(SampleID[i],2), log10x.exp.IQ[i,], lwd=3, lend = 1, col="black")
	  }

	###
	# Tables to print
	###
	## ISE parameter estimates
	cat("\n\nNon-linear parameter estimates and 95% CIs for\n\t y = a + b log(x + c)\n")
	for (i in 1:Num.ISEs) {
		cat("\nISE #",i,":\n", sep="")
		table1 = format(cbind(c(ahat[i], bhat[i], chat[i], sigmahat[i]), 
			c(ahat.lcl[i], bhat.lcl[i], chat.lcl[i], sigmahat.lcl[i]), 
			c(ahat.ucl[i], bhat.ucl[i], chat.ucl[i], sigmahat.ucl[i])), nsmall=2, digits=3)
		colnames(table1)= c("Parameter estimate", "Lower limit", "Upper limit")
		rownames(table1)= c("a", "b", "c", "sigma")
		print(table1, quote=F)
		cat("\nEstimated (log) LOD (S/N=3): ", format(LOD.SN3[i], digits=3),"\n\n")
	}

	## Sample concentrations
	SampleID = seq(data$M)
	table2 = cbind(SampleID, format(log10x.exp, nsmall=2, digits=3))
	cat("\nBayesian calibration of experimental samples:\n\n")
	rownames(table2)= rep("", data$M)
	print(table2, quote=F)

	#   If flag.datareplicated == 1, the sample was replicated
	#   for the BUGs model, and the the user is told
	#   that Sample 1 and Sample 2 are the same 
      if(flag.datareplicated==1) {
		cat("\nNote: Sample 1 and Sample 2 are the same sample.\n")
		cat("The sample was duplicated for computational reasons.\n\n")
	}
	
	###
	# Relevant output
	###
	return(list(log10x.exp=log10x.exp, log10x.exp.IQ= log10x.exp.IQ, 
		ahat=ahat, bhat=bhat, chat=chat, cstarhat=cstarhat, sigmahat=sigmahat, LOD.SN3=LOD.SN3, 
		ahat.lcl=ahat.lcl, ahat.ucl=ahat.ucl, bhat.lcl=bhat.lcl, bhat.ucl=bhat.ucl, 
		chat.lcl=chat.lcl, chat.ucl=chat.ucl, sigmahat.lcl=sigmahat.lcl, sigmahat.ucl=sigmahat.ucl,
		LOD.SN3hat = LOD.SN3hat, LOD.SN3.lcl = LOD.SN3.lcl, LOD.SN3.ucl = LOD.SN3.ucl,
		LOD.SN3.Q1 = LOD.SN3.Q1, LOD.SN3.Q3 = LOD.SN3.Q3))	
}


### Loads data for Multiple ISEs from tab delimited files and puts them in an appropriate format.
## One file should contain calibration data, with a header row as follows:
# ISEID: A number (1, 2, 3, ...; must be consecutive) identifying the ISE
# log10x: The known concentration on a log10 scale
# emf: The measured emf
#
## The second file should contain experimental data (or set calibration.only=T if there is no experimental data), with a header row as follows:
# For the basic model (stdadd = F):
# 	ISEID: A number (1, 2, 3, ...; must be consecutive) identifying the ISE
# 	SampleID: A number identifying the unknown sample that is being measured
# 	emf: The measured emf
# For the standard addition model (stdadd = T):
# 	ISEID: A number (1, 2, 3, ...; must be consecutive) identifying the ISE
# 	SampleID: A number identifying the unknown sample that is being measured
# 	emf1: The measured emf for the sample
#	emf2: The measured emf for the sample plus the standard addition
# 	V.s: volume of the sample (any convenient unit, e.g. ml or l)
# 	V.add: volume of the addition (same units as V.s)
# 	conc.add: concentration of the addition (mol/l)
#
load.ISE.data = function(filename.experimental = NA, filename.calibration, stdadd = F, calibration.only=F) {
   if (!calibration.only) {
	# Load the experimental and calibration files
	data.exp = read.delim(filename.experimental,
		header = TRUE, sep = "\t", quote="\"", dec=".", fill = TRUE, comment.char="")
	
	data.calib = read.delim(filename.calibration,
		header = TRUE, sep = "\t", quote="\"", dec=".", fill = TRUE, comment.char="")

	print("Calibration data")
	print(data.calib)
	cat("\n")
	print("Experimental data")
	print(data.exp)

	# Format data from the calibration file
	N = nrow(data.calib)
	R = max(data.calib$ISEID)
	ISEID = data.calib$ISEID
	log10x = data.calib$log10x
	emf = data.calib$emf

	# Plot the calibration data, four ISEs at a time
	par(mfrow=c(2,2))
	for (i in 1:R) {
		plot(x = log10x[ISEID==i], y = emf[ISEID==i], 
			xlab=expression(paste(log[10], " { ", italic(x), " }" )), 
			ylab="emf", main=paste("ISE #", i, sep=""))
	}
	

	# Format data from the experimental file
	M = max(data.exp$SampleID)
	M.obs = nrow(data.exp)
	ISEID.exp = data.exp$ISEID
	xID.exp = data.exp$SampleID
	if (stdadd == T) {
		delta.emf = data.exp$emf2 - data.exp$emf1
		V.s = data.exp$V.s
		V.add = data.exp$V.add
		conc.add = data.exp$conc.add
	}
	if (stdadd != T) {
		emf.exp = data.exp$emf
	}

	### If there is only one ISE, format the data accordingly:
	if(R == 1) {
		if (stdadd == T) {
			data.out = list(N=N, log10x = log10x, emf=emf,
				M=M, delta.emf = delta.emf, 
				V.s = V.s, V.add = V.add, conc.add = conc.add)
		}
		if (stdadd != T) {
			data.out = list(N=N, log10x = log10x, emf=emf,
				M=M, emf.exp = emf.exp)
		}
	}

	### If multiple ISEs, format as follows:
	if(R > 1) {
		if (stdadd == T) {
			data.out = list(N=N, R=R, ISEID=ISEID, log10x = log10x, emf=emf,
				M=M, M.obs = M.obs, ISEID.exp = ISEID.exp,
				xID.exp=xID.exp, delta.emf = delta.emf, 
				V.s = V.s, V.add = V.add, conc.add = conc.add)
		}
		if (stdadd != T) {
			data.out = list(N=N, R=R, ISEID=ISEID, log10x = log10x, emf=emf,
				M=M, M.obs = M.obs, ISEID.exp = ISEID.exp,
				xID.exp=xID.exp, emf.exp = emf.exp)
		}
	}
   }
   if (calibration.only) {
	data.out = load.ISE.data.calibration(filename.calibration = filename.calibration)
   }
   return(data.out)   

}

##
# A sub-function called by load.ISE.data for when there is only calibration data (i.e. the goal is to characterise an ISE, not estimate
#   activity for unknowns
##

load.ISE.data.calibration = function(filename.calibration) {
	# Load the calibration files
	data.calib = read.delim(filename.calibration,
		header = TRUE, sep = "\t", quote="\"", dec=".", fill = TRUE, comment.char="")

	print("Calibration data")
	print(data.calib)
	cat("\n")

	# Format data from the calibration file
	N = nrow(data.calib)
	R = max(data.calib$ISEID)
	ISEID = data.calib$ISEID
	log10x = data.calib$log10x
	emf = data.calib$emf

	# Plot the calibration data, four ISEs at a time
	par(mfrow=c(2,2))
	for (i in 1:R) {
		plot(x = log10x[ISEID==i], y = emf[ISEID==i], 
			xlab=expression(paste(log[10], " { ", italic(x), " }" )), 
			ylab="emf", main=paste("ISE #", i, sep=""))
	}
	
	if (R == 1) { data.out = list(N=N, log10x = log10x, emf=emf) }
	if (R > 1) { data.out = list(N=N, R=R, ISEID=ISEID, log10x = log10x, emf=emf) }

	return(data.out)
}


###
# Calculates calibration curve to characterise an ISE
# Required inputs are:
#  data:        Data formatted using load.ISE.data
#  model.path:  The directory where the BUGS model is located (e.g. "c:/myisefolder")
#  model.name:  The name of the BUGS model (e.g. "Single ISE model Calibration only.txt")
#  single.ISE:  T or F, indicating whether the data come from a single ISE or multiple ISEs
#
# Technical inputs that do not require specification: 
#  burnin:      The number of initial parameter estimates to discard while tuning the system
#  iters:       The total number of iterations
#  chains:      The number of parallel chains to run (useful for diagnostics)
#  ### Note:    The final estimates are based on (iters - burnin)*chains number of simulations
#  a.init:      Initial value passed to gen.inits.single (scalar) or gen.inits.multiple (vector)
#  b.init:      Similar
#  cstar.init:  Similar
#  logc.limits: Upper and lower climits for log(c) (defaults to -8.9 and -1.9)
#  sigma.upper: Upper limit for sigma.init (recommended to include a value)
#  offset:	The initial value for the slope uses the last and the last - offset point (defaults to 1).
#  ### Note:    Default values may be altered if there are convergence problems in the 
#               non-linear regression
#  diagnostic.print:  Prints information used to determine convergence
###

characterise.ise <- function(data, model.path, model.name, single.ISE=NA, Z=NA, temperature = 21, 
	burnin=25000, iters = 50000, chains=4, thin = 1,
	a.init= NA, b.init=NA, cstar.init=NA, logc.limits = c(-8.9, -1.9), sigma.upper = NA, diagnostic.print=F, 
	offset = 1, keep.coda=F, coda.n=NA) {

	# Location of the BUGs model
	ISE.model <- file.path(model.path, model.name)

	# Error checking
	if (is.na(single.ISE)) {
		print("Please specify single.ISE=T or single.ISE=F to indicate a single or multiple ISEs")
		return()
	}

	# Error checking
	if (is.na(Z)) {
		print("Please specify the ionic valence using Z= (e.g. bayesian.calibrate(..., Z = 2,...)")
		return()
	}

	# Initial values
	cat("\nGenerating initial values...\n")
	if (single.ISE==T) {
		# Number of ISEs = 1
		Num.ISEs = 1
		# Generate initial values
		inits <- rep(list(NA), chains)
		for (i in 1:chains) {
			inits[[i]] <- gen.inits.single(data, a.init=a.init, b.init=b.init,
				cstar.init= cstar.init, logc.limits=logc.limits, sigma.upper=sigma.upper, stdadd = F, offset=offset, calibration.only=T)
		}
	}
	if (single.ISE==F) {
		# Generate initial values
		inits <- rep(list(NA), chains)
		# Number of ISEs given in data
		Num.ISEs = data$R
		for (i in 1:chains) {
			inits[[i]] <- gen.inits.multiple(data, a.init=a.init, b.init=b.init, 
				cstar.init= cstar.init, logc.limits=logc.limits, sigma.upper= sigma.upper, stdadd = F, offset=offset, calibration.only=T)
		}
	}
	# Parameters to monitor
	parameters <- c("a", "b", "c", "cstar", "sigma")

	# Calculate theoretical slope at mV level
	data$mu.b = 1000*log(10)*8.31433*(temperature + 273.15)/(96487*Z)

	# Call BUGS
	cat("Running Bayesian model...\n\n")
	ISE.Bayes <- bugs(data, inits, parameters, ISE.model,
		n.chains = chains, n.iter = iters, n.burnin=burnin, n.thin=thin,
		program = "openbugs", working.directory = NULL, digits=8)

	# Print and plot diagnostics
	if(diagnostic.print==T) {
		print(ISE.Bayes, digits=9)
		plot(ISE.Bayes)
	}

  if(single.ISE==T) {
	# Other vars
	ahat.coda = as.vector(ISE.Bayes$sims.array[,,1])
	coda.length = length(ahat.coda)
	bhat.coda = as.vector(ISE.Bayes$sims.array[,,2])
	cstarhat.coda = as.vector(ISE.Bayes$sims.array[,,4])
	chat.coda = cstarhat.coda^10
	sigmahat.coda = as.vector(ISE.Bayes$sims.array[,,5])
	ahat = median(ahat.coda)
	bhat = median(bhat.coda)
	cstarhat = median(cstarhat.coda)
	chat = cstarhat^10
	sigmahat = median(sigmahat.coda)

	ahat.lcl = quantile(ahat.coda, 0.025)
	bhat.lcl = quantile(bhat.coda, 0.025)
	cstarhat.lcl = quantile(cstarhat.coda, 0.025)
	chat.lcl = cstarhat.lcl^10
	ahat.ucl = quantile(ahat.coda, 0.975)
	bhat.ucl = quantile(bhat.coda, 0.975)
	cstarhat.ucl = quantile(cstarhat.coda, 0.975)
	chat.ucl = cstarhat.ucl^10
	sigmahat.lcl = quantile(sigmahat.coda, 0.025)
	sigmahat.ucl = quantile(sigmahat.coda, 0.975)


	# Derived variable
	LOD.SN3 = log10( cstarhat^10 *(10^(3*sigmahat/abs(bhat))-1))

	LOD.SN3.coda = 10*log10(cstarhat.coda) + log10(10^(3*sigmahat.coda /abs( bhat.coda ) ) - 1)
	LOD.SN3hat = median(LOD.SN3.coda)
	LOD.SN3.Q1 = quantile(LOD.SN3.coda, 0.25)
	LOD.SN3.Q3 = quantile(LOD.SN3.coda, 0.75)
	LOD.SN3.lcl = quantile(LOD.SN3.coda, 0.025)
	LOD.SN3.ucl = quantile(LOD.SN3.coda, 0.975)

  }

  if (single.ISE == F) {	
	# Other vars
	coda.length = length(as.vector(ISE.Bayes$sims.array[,,1]))
	ahat.coda = matrix(NA, nrow = coda.length, ncol = data$R)
	bhat.coda <- cstarhat.coda <- chat.coda <- sigmahat.coda <- LOD.SN3.coda <- ahat.coda
	
	ahat = rep(NA, data$R)
	ahat.lcl = rep(NA, data$R)
	ahat.ucl = rep(NA, data$R)
	bhat = rep(NA, data$R)
	bhat.lcl = rep(NA, data$R)
	bhat.ucl = rep(NA, data$R)
	chat = rep(NA, data$R)
	chat.lcl = rep(NA, data$R)
	chat.ucl = rep(NA, data$R)
	cstarhat = rep(NA, data$R)
	cstarhat.lcl = rep(NA, data$R)
	cstarhat.ucl = rep(NA, data$R)
	sigmahat = rep(NA, data$R)
	sigmahat.lcl = rep(NA, data$R)
	sigmahat.ucl = rep(NA, data$R)
	LOD.SN3 = rep(NA, data$R)

	LOD.SN3hat = rep(NA, data$R)
	LOD.SN3.lcl = rep(NA, data$R)
	LOD.SN3.ucl = rep(NA, data$R)
	LOD.SN3.Q1 = rep(NA, data$R)
	LOD.SN3.Q3 = rep(NA, data$R)

	for (j in 1:data$R) {
		ahat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,j])
		bhat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,data$R + j])
		cstarhat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,3*data$R + j])
		sigmahat.coda[,j] = as.vector(ISE.Bayes$sims.array[,,4*data$R + j])
		chat.coda[,j] = cstarhat.coda[,j]^10

		ahat.q = quantile(ahat.coda[,j], c(0.025, 0.5, 0.975))
		ahat[j] = ahat.q[2]
		ahat.lcl[j] = ahat.q[1]
		ahat.ucl[j] = ahat.q[3]
		bhat.q = quantile(bhat.coda[,j], c(0.025, 0.5, 0.975))
		bhat[j] = bhat.q[2]
		bhat.lcl[j] = bhat.q[1]
		bhat.ucl[j] = bhat.q[3]
		chat.q = quantile(chat.coda[,j], c(0.025, 0.5, 0.975))
		chat[j] = chat.q[2]
		chat.lcl[j] = chat.q[1]
		chat.ucl[j] = chat.q[3]
		cstarhat.q = quantile(cstarhat.coda[,j], c(0.025, 0.5, 0.975))
		cstarhat[j] = cstarhat.q[2]
		cstarhat.lcl[j] = cstarhat.q[1]
		cstarhat.ucl[j] = cstarhat.q[3]
		sigmahat.q = quantile(sigmahat.coda[,j], c(0.025, 0.5, 0.975))
		sigmahat[j] = sigmahat.q[2]
		sigmahat.lcl[j] = sigmahat.q[1]
		sigmahat.ucl[j] = sigmahat.q[3]

		# Derived variable
		LOD.SN3[j] = 10*log10( cstarhat[j]) + log10(10^(3*sigmahat[j]/abs(bhat[j]))-1)

		LOD.SN3.coda[,j] = 10*log10( cstarhat.coda[,j]) + log10( 10 ^ (3*sigmahat.coda[,j] / abs( bhat.coda[,j] )) - 1)
		LOD.SN3.q = quantile(LOD.SN3.coda[,j], c(0.025, 0.25, 0.5, 0.75, 0.975))
		LOD.SN3hat[j] = LOD.SN3.q[3]
		LOD.SN3.lcl[j] = LOD.SN3.q[1]
		LOD.SN3.ucl[j] = LOD.SN3.q[5]
		LOD.SN3.Q1[j] = LOD.SN3.q[2]
		LOD.SN3.Q3[j] = LOD.SN3.q[3]
   	}	
   }


	###
	# Tables to print
	###
	## ISE parameter estimates
	cat("\n\nNon-linear parameter estimates and 95% CIs for\n\t y = a + b log(x + c)\n")
	for (i in 1:Num.ISEs) {
		cat("\nISE #",i,":\n", sep="")
		table1 = format(cbind(c(ahat[i], bhat[i], chat[i], sigmahat[i]), 
			c(ahat.lcl[i], bhat.lcl[i], chat.lcl[i], sigmahat.lcl[i]), 
			c(ahat.ucl[i], bhat.ucl[i], chat.ucl[i], sigmahat.ucl[i])), nsmall=2, digits=3)
		colnames(table1)= c("Parameter estimate", "Lower limit", "Upper limit")
		rownames(table1)= c("a", "b", "c", "sigma")
		print(table1, quote=F)
		cat("\nEstimated (log) LOD (S/N=3): ", format(LOD.SN3[i], digits=3),"\n\n")

	}
	
	###
	# Relevant output
	###
	if (keep.coda) {
  	   if (!is.na(coda.n)) {
		if (coda.n > coda.length) {
			cat("\nWarning: requested coda sub-sample is longer than coda. Full coda returned.\n")
			coda.n = coda.length
		}
		coda.sample = sample(seq(coda.length), coda.n, replace=F) 
		if (single.ISE == T) {
			ahat.coda = ahat.coda[coda.sample]
			bhat.coda = bhat.coda[coda.sample]
			chat.coda = chat.coda[coda.sample]
			cstarhat.coda = cstarhat.coda[coda.sample]
			sigmahat.coda = sigmahat.coda[coda.sample]
			LOD.SN3.coda = LOD.SN3.coda[coda.sample]
		}
		if (single.ISE == F) {
			ahat.coda = ahat.coda[coda.sample,]
			bhat.coda = bhat.coda[coda.sample,]
			chat.coda = chat.coda[coda.sample,]
			cstarhat.coda = cstarhat.coda[coda.sample,]
			sigmahat.coda = sigmahat.coda[coda.sample,]
			LOD.SN3.coda = LOD.SN3.coda[coda.sample,]
		}
	   }
	   return(list(ahat.coda=ahat.coda, bhat.coda=bhat.coda, chat.coda = chat.coda, sigmahat.coda=sigmahat.coda, cstarhat.coda=cstarhat.coda, LOD.SN3.coda=LOD.SN3.coda, 
		ahat=ahat, bhat=bhat, chat=chat, cstarhat=cstarhat, sigmahat=sigmahat, LOD.SN3=LOD.SN3, 
		ahat.lcl=ahat.lcl, ahat.ucl=ahat.ucl, bhat.lcl=bhat.lcl, bhat.ucl=bhat.ucl, 
		chat.lcl=chat.lcl, chat.ucl=chat.ucl, cstarhat.lcl=cstarhat.lcl, cstarhat.ucl=cstarhat.ucl, sigmahat.lcl=sigmahat.lcl, sigmahat.ucl=sigmahat.ucl,
		LOD.SN3hat = LOD.SN3hat, LOD.SN3.lcl = LOD.SN3.lcl, LOD.SN3.ucl = LOD.SN3.ucl,
		LOD.SN3.Q1 = LOD.SN3.Q1, LOD.SN3.Q3 = LOD.SN3.Q3))	
	}
	if (!keep.coda) {
	   return(list(
		ahat=ahat, bhat=bhat, chat=chat, cstarhat=cstarhat, sigmahat=sigmahat, LOD.SN3=LOD.SN3, 
		ahat.lcl=ahat.lcl, ahat.ucl=ahat.ucl, bhat.lcl=bhat.lcl, bhat.ucl=bhat.ucl, 
		chat.lcl=chat.lcl, chat.ucl=chat.ucl, cstarhat.lcl=cstarhat.lcl, cstarhat.ucl=cstarhat.ucl, sigmahat.lcl=sigmahat.lcl, sigmahat.ucl=sigmahat.ucl,
		LOD.SN3hat = LOD.SN3hat, LOD.SN3.lcl = LOD.SN3.lcl, LOD.SN3.ucl = LOD.SN3.ucl,
		LOD.SN3.Q1 = LOD.SN3.Q1, LOD.SN3.Q3 = LOD.SN3.Q3))	
	}
}

characterize.ise <- characterise.ise


plot.samples = function(data, output, plot.limits = c(-15, 0), x.ticks = NA, y.ticks = NA,
	my.ylab = expression(paste(log[10], " { ", italic(x), " }" )), my.xlab = "Sample ID") {
	### Plot calibrated samples
	par(mfrow=c(2,2))
	par(mar=c(4,4.0,1,0)+0.1)
	par(mgp=c(2, 0.5, 0), tcl = -0.2)

	SampleID = seq(data$M)
	delta = 0.015*data$M
	S.low = SampleID - delta
	S.high = SampleID + delta
	S.llow = SampleID - 0.5*delta
	S.lhigh = SampleID + 0.5*delta

	log10x.exp = output$log10x.exp
	log10x.exp.IQ = output$log10x.exp.IQ

	plot(rep(SampleID[1],2), c(log10x.exp[1,2], log10x.exp[1,3]), type="l", axes=F,
		xlim=c(min(SampleID), max(SampleID)), ylim=plot.limits, lend = 1, 
		xlab = my.xlab, ylab=my.ylab
	)
	box()
	if(!is.na(x.ticks[1])) {
		axis(1, at = x.ticks, cex.axis= 0.8)
	}
	if(is.na(x.ticks[1])) {
		axis(1, at = SampleID, cex.axis = 0.8)
	}
	if(is.na(y.ticks[1])) {
		axis(2, cex.axis= 0.8)
	}
	if(!is.na(y.ticks[1])) {
		axis(2, at = y.ticks, cex.axis= 0.8)
	}

	for (i in 1:data$M) {
		lines(rep(SampleID[i],2), c(log10x.exp[i,2], log10x.exp[i,3]), col="black")
		lines(c(S.llow[i], S.lhigh[i]), rep(log10x.exp[i,2], 2) ,  col="black")
		lines(c(S.llow[i], S.lhigh[i]), rep(log10x.exp[i,3], 2), col="black")
		lines(c(S.low[i], S.high[i]), rep(log10x.exp[i,1], 2), lwd=1, col="black")
		lines(rep(SampleID[i],2), log10x.exp.IQ[i,], lwd=3, lend = 1, col="black")
	 }

}
